import pyvisa
import pylabframe
from enum import Enum

_visa_rm = pyvisa.ResourceManager()
_connected_devices = {}


def get_device(id):
    if id in _connected_devices:
        return _connected_devices[id]
    else:
        dev = _connect_device(id)
        _connected_devices[id] = dev
        return dev


def _connect_device(id):
    from . import drivers
    hw_conf = read_hw_config()
    device_parts = hw_conf[id].split("@")
    device_class = device_parts[0]
    device_address = "@".join(device_parts[1:])

    device_class = eval(f"drivers.{device_class}")

    dev = device_class(id, device_address)

    return dev


def read_hw_config(computer_name=None):
    if computer_name is None:
        computer_name = pylabframe.general.get_computer_name()

    return pylabframe.general.load_config_file(f"hardware/{computer_name}")


class Device:
    def __init__(self, id, error_on_double_connect=True, **kw):
        if id in _connected_devices and error_on_double_connect:
            raise RuntimeError(f"Device {id} already connected")

    @classmethod
    def list_available(cls):
        return []


class VisaDevice(Device):
    def __init__(self, id, address, **kw):
        super().__init__(id, **kw)
        self.address = address
        self.visa_instr: pyvisa.resources.messagebased.MessageBasedResource = _visa_rm.open_resource(address)

    def __del__(self):
        self.visa_instr.close()
        del self.visa_instr

    @classmethod
    def list_available(cls):
        return list(_visa_rm.list_resources())

    def get_identifier(self, sanitize=True):
        response = self.visa_instr.query("*IDN?")
        if sanitize:
            response = response.strip()
        return response

    def wait_until_done(self):
        self.visa_instr.write("*OPC?")
        is_done = False
        while not is_done:
            try:
                self.visa_instr.read()
                is_done = True
            except pyvisa.VisaIOError as e:
                if e.error_code != pyvisa.constants.StatusCode.error_timeout:
                    # re-raise anything other than a time-out
                    raise e

    @classmethod
    def visa_property(cls, visa_cmd: str, read_only=False, read_conv=str, write_conv=str):
        def visa_getter(self: VisaDevice):
            # doing this gives us access to object properties (eg channel id) that can be put in the command string
            fmt_visa_cmd = visa_cmd
            if hasattr(self, "query_params"):
                fmt_visa_cmd = fmt_visa_cmd.format(**self.query_params)
            response = self.visa_instr.query(f"{fmt_visa_cmd}?")
            response = read_conv(response.strip())
            return response

        if not read_only:
            def visa_setter(self: VisaDevice, value):
                fmt_visa_cmd = visa_cmd
                if hasattr(self, "query_params"):
                    fmt_visa_cmd = fmt_visa_cmd.format(**self.query_params)
                cmd = f"{fmt_visa_cmd} {write_conv(value)}"
                self.visa_instr.write(cmd)
        else:
            visa_setter = None

        prop = property(visa_getter, visa_setter)
        return prop


def enum_conv(e, allow_other_types=True):
    if isinstance(e, Enum):
        return e.value
    elif allow_other_types:
        return e
    else:
        raise ValueError(f"{e} is not an enum type")

def str_conv(s):
    return s.replace('"', '')


import sys


def get_computer_name():
    cname = None
    with open("pylabframe/config/_computer_name.cfg", "r") as f:
        cname = f.read().strip()

    return cname


def load_config_file(filename):
    config_content = {}
    if not filename.endswith(".cfg"):
        filename += ".cfg"
    with open(f"pylabframe/config/{filename}", "r") as f:
        lines = f.readlines()

    for line in lines:
        if line.strip().startswith("#"):
            # this is a comment, continue
            continue
        if line.strip() == '':
            # empty line, continue
            continue

        # if "#" in line:
        #     # comment somewhere on the line
        #     hash_idx = line.find("#")
        #     line = line[:hash_idx]

        if "=" not in line:
            raise ValueError(f"Invalid config line {line}")

        split_idx = line.find("=")
        k = line[:split_idx].strip()
        v = line[split_idx+1:].strip()

        config_content[k] = v

    return config_content


import numpy as np

class NumericalData:
    def __init__(self, data_array=None, x_axis=None, y_axis=None, axes=None, metadata=None, convert_to_numpy=True):
        if convert_to_numpy:
            data_array = np.asarray(data_array)
        self.data_array = data_array
        self.axes = axes if axes is not None else []
        self.metadata = metadata if metadata is not None else {}
        if x_axis is not None:
            self.set_axis(0, x_axis, convert_to_numpy=convert_to_numpy)
        if y_axis is not None:
            self.set_axis(1, y_axis, convert_to_numpy=convert_to_numpy)

    def set_axis(self, ax_index, ax_values, convert_to_numpy=True):
        if len(self.axes) < ax_index + 1:
            self.axes = self.axes + [None] * (ax_index + 1 - len(self.axes))
        if convert_to_numpy:
            ax_values = np.asarray(ax_values)
        self.axes[ax_index] = ax_values

    @property
    def x_axis(self):
        return self.axes[0]
    @x_axis.setter
    def x_axis(self, ax_values):
        self.set_axis(0, ax_values)

    @property
    def y_axis(self):
        return self.axes[1]
    @y_axis.setter
    def y_axis(self, ax_values):
        self.set_axis(1, ax_values)

    @property
    def z_axis(self):
        return self.axes[2]
    @z_axis.setter
    def z_axis(self, ax_values):
        self.set_axis(2, ax_values)

    # patch all calls that don't work on the NumericalData object directly through to the underlying data_array
    def __getattr__(self, item):
        return getattr(self.data_array, item)

    # @property
    # def ndim(self):
    #     return self.data_array.ndim
    #
    # @property
    # def shape(self):
    #     return self.data_array.shape

    def plot(self, plot_axis=None, x_label=None, y_label=None, auto_label=True, **kw):
        if self.ndim == 1:
            return self.plot_1d(plot_axis, x_label=None, y_label=None, auto_label=auto_label, **kw)
        elif self.ndim == 2:
            return self.plot_2d(plot_axis, x_label=None, y_label=None, auto_label=auto_label, **kw)
        else:
            raise NotImplementedError(f"No plotting method available for {self.ndim}-dimensional data")

    def plot_1d(self, plot_axis=None, x_label=None, y_label=None, auto_label=True, **kw):
        # set some defaults
        if 'm' not in kw and 'marker' not in kw:
            kw['marker'] = '.'

        if plot_axis is None:
            import matplotlib.pyplot as plt
            plot_axis = plt.gca()

        plot_axis.plot(self.x_axis, self.data_array, **kw)
        if x_label is None and auto_label and "x_label" in self.metadata:
            x_label = self.metadata["x_label"]
            if "x_unit" in self.metadata:
                x_label += f" ({self.metadata['x_unit']})"
        if x_label is not None:
            plot_axis.set_xlabel(x_label)
        if y_label is None and auto_label and "y_label" in self.metadata:
            y_label = self.metadata["y_label"]
            if "y_unit" in self.metadata:
                y_label += f" ({self.metadata['y_unit']})"
        if y_label is not None:
            plot_axis.set_ylabel(y_label)




import numpy as np

from . import device
from enum import Enum
from .device import enum_conv, str_conv
import pylabframe.data

# helper definition
visa_property = device.VisaDevice.visa_property


class TektronixScope(device.VisaDevice):
    NUM_CHANNELS = 2

    class RunModes(Enum):
        CONTINUOUS = "RUNST"
        SINGLE = "SEQ"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # initialize channels
        self.channels: list[TektronixScope.Channel] = [self.Channel(i+1, self) for i in range(self.NUM_CHANNELS)]

    # global scpi properties
    record_length = visa_property("horizontal:recordlength", read_conv=int, write_conv=int)
    run_mode = visa_property("acquire:stopafter", read_conv=RunModes, write_conv=enum_conv)
    run_state = visa_property("acquire:state", read_conv=bool, write_conv=int)
    x_scale = visa_property("horizontal:scale", read_conv=float, write_conv=float)

    # waveform transfer properties
    waveform_points = visa_property("wfmoutpre:nr_pt", read_only=True, read_conv=int)
    waveform_y_multiplier = visa_property("wfmoutpre:ymult", read_only=True, read_conv=float)
    waveform_y_offset_levels = visa_property("wfmoutpre:yoff", read_only=True, read_conv=float)
    waveform_y_zero = visa_property("wfmoutpre:yzero", read_only=True, read_conv=float)
    waveform_y_unit = visa_property("wfmoutpre:yunit", read_only=True, read_conv=str_conv)

    waveform_x_increment = visa_property("wfmoutpre:xincr", read_only=True, read_conv=float)
    waveform_x_zero = visa_property("wfmoutpre:xzero", read_only=True, read_conv=float)
    waveform_x_unit = visa_property("wfmoutpre:xunit", read_only=True, read_conv=str_conv)

    def initialize_waveform_transfer(self, channel_id, start=1, stop=None):
        self.visa_instr.write(f"data:source ch{channel_id}")
        self.visa_instr.write(f"data:start {start}")
        if stop is None:
            # default to full waveform
            stop = self.record_length
        self.visa_instr.write(f"data:stop {stop}")
        self.visa_instr.write("data:encdg fast")
        self.visa_instr.write("data:width 2")
        self.visa_instr.write("header 0")

    def do_waveform_transfer(self):
        wfm_raw = self.visa_instr.query_binary_values("curve?", datatype='h', is_big_endian=True, container=np.array)
        wfm_converted = (wfm_raw * self.waveform_y_multiplier) + self.waveform_y_zero
        time_axis = (np.arange(self.waveform_points) * self.waveform_x_increment) + self.waveform_x_zero

        metadata = {
            "x_unit": self.waveform_x_unit,
            "x_label": f"time",
            "y_unit": self.waveform_y_unit,
            "y_label": f"signal",
        }
        data_obj = pylabframe.data.NumericalData(wfm_converted, x_axis=time_axis, metadata=metadata)
        return data_obj

    def get_channel_waveform(self, channel_id, start=1, stop=None):
        self.initialize_waveform_transfer(channel_id, start=start, stop=stop)
        wfm = self.do_waveform_transfer()
        return wfm

    # channel properties
    class Channel:
        def __init__(self, channel, device):
            self.channel_id = channel
            self.query_params = {'channel_id':  channel}
            self.device: "TektronixScope" = device
            self.visa_instr = self.device.visa_instr

        y_scale = visa_property("ch{channel_id}:scale", read_conv=float, write_conv=float)
        offset = visa_property("ch{channel_id}:offset", read_conv=float, write_conv=float)
        termination = visa_property("ch{channel_id}:termination", read_conv=float, write_conv=float)
        inverted = visa_property("ch{channel_id}:invert", read_conv=bool, write_conv=int)

        mean = visa_property("measu:meas{channel_id}:mean", read_only=True, read_conv=float)

        def get_waveform(self, start=1, stop=None):
            return self.device.get_channel_waveform(self.channel_id, start=start, stop=stop)


